import os
from OpenGL.GL import *
from PenguinMol3D.materials.base_material import BaseMaterial

"""
   This module contains material class, which uses a classic Blinn-Phong lighting model. 
   You can create your own materials by subclassing it and configuring 
   'shininess' and 'specular strength' uniform values to get unique visual appearance
   
   You can get Shininess of different Phong materials following this link:
   http://web.archive.org/web/20100725103839/http://www.cs.utk.edu/~kuck/materials_ogl.htm
   
   NOTE: to get comparable result with Blinn-Phong lighting model this values should be increased in 2-4 fold.
   
   Also, check 'phong_materials' module for ready to use presets
"""

class PhongMaterial(BaseMaterial):
    def __init__(self, properties={}):

        vert_shader_name = "base.vert"
        if properties["use_instanced_rendering"]:
            vert_shader_name = "instanced.vert"

        vert_shader_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                        "shaders",
                                        vert_shader_name)

        frag_shader_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                        "shaders",
                                        "blinn_phong_material.frag")

        BaseMaterial.__init__(self, vert_shader_path, frag_shader_path)

        if not properties["use_instanced_rendering"]:
            self.add_uniform(None, "mat4", "model_matrix")

        self.add_uniform(False, "bool", "use_vertex_colors")
        self.add_uniform([1., 1., 1.], "vec3", "base_color")
        self.add_uniform([0., 0., 0.], "vec3", "view_position")

        self.add_uniform(1., "float", "specular_strength")
        self.add_uniform(20., "float", "shininess")

        self.locate_uniforms()

        self._settings["cull_face"] = False
        self._settings["front_side"] = True
        self._settings["back_side"] = True
        self._settings["wireframe"] = False
        self._settings["line_width"] = 1

        self.set_properties(properties)
        self.update_render_settings()

    def update_render_settings(self):

        if not self._settings["cull_face"]:
            glDisable(GL_CULL_FACE)
        else:
            glEnable(GL_CULL_FACE)

        if not self._settings["front_side"]:
            glCullFace(GL_FRONT)

        if not self._settings["back_side"]:
            glCullFace(GL_BACK)

        if self._settings["wireframe"]:
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        else:
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)

        glLineWidth(self._settings["line_width"])



from PenguinMol3D.materials import (
    phong_material as pm,
    pbr_material as pb
)


class BaseFactory:
    def __init__(self,
                 material_type: pm.PhongMaterial | pb.PBRMaterial = pb.PBRMaterial,
                 color_scale: float = 1.0):

        self._material_type = material_type
        self._color_scale = color_scale

    @property
    def material_type(self):
        return self._material_type

    @property
    def color_scale(self):
        return self._color_scale


from PenguinMol3D.materials.phong_material import PhongMaterial

"""
   This module contains presets of various materials. By manipulating with material shininess 
   and specular strength we can configure the reflectiveness of surface mimicking properties 
   of real world materials. This, however, has no impact on material color under different 
   types of lighting, which depends on the atom type.
"""

class BrassMaterial(PhongMaterial):
    def __init__(self, properties: dict = {}):
        PhongMaterial.__init__(self, properties)

        self.uniforms["specular_strength"].data = 1.
        self.uniforms["shininess"].data = 83.6922   #Phong 27.8974


class BronzeMaterial(PhongMaterial):
    def __init__(self, properties: dict = {}):
        PhongMaterial.__init__(self, properties)

        self.uniforms["specular_strength"].data = 1.
        self.uniforms["shininess"].data = 76.8       #Phong 25.6


class ChromeMaterial(PhongMaterial):
    def __init__(self, properties: dict = {}):
        PhongMaterial.__init__(self, properties)

        self.uniforms["specular_strength"].data = 1.
        self.uniforms["shininess"].data = 230.4      #Phong 76.8


class CopperMaterial(PhongMaterial):
    def __init__(self, properties: dict = {}):
        PhongMaterial.__init__(self, properties)

        self.uniforms["specular_strength"].data = 1.
        self.uniforms["shininess"].data = 38.4       #Phong 51.2


class GoldMaterial(PhongMaterial):
    def __init__(self, properties: dict = {}):
        PhongMaterial.__init__(self, properties)

        self.uniforms["specular_strength"].data = 1.
        self.uniforms["shininess"].data = 153.6      #Phong 51.2


class ObsidianMaterial(PhongMaterial):
    def __init__(self, properties: dict = {}):
        PhongMaterial.__init__(self, properties)

        self.uniforms["specular_strength"].data = 1.
        self.uniforms["shininess"].data = 115.2      #Phong 38.4


class PearlMaterial(PhongMaterial):
    def __init__(self, properties: dict = {}):
        PhongMaterial.__init__(self, properties)

        self.uniforms["specular_strength"].data = 1.
        self.uniforms["shininess"].data = 33.792     #Phong 11.264


class PlasticMaterial(PhongMaterial):
    def __init__(self, properties: dict = {}):
        PhongMaterial.__init__(self, properties)

        self.uniforms["specular_strength"].data = 1.
        self.uniforms["shininess"].data = 96.        #Phong 32.


class RubberMaterial(PhongMaterial):
    def __init__(self, properties: dict = {}):
        PhongMaterial.__init__(self, properties)

        self.uniforms["specular_strength"].data = 1.
        self.uniforms["shininess"].data = 30.         #Phong 10.



class Materials:
    Brass = BrassMaterial
    Bronze = BronzeMaterial
    Chrome = ChromeMaterial
    Copper = CopperMaterial
    Gold = GoldMaterial
    Obsidian = ObsidianMaterial
    Pearl = PearlMaterial
    Plastic = PlasticMaterial
    Rubber = RubberMaterial



"""PenguinMol3D: 3D visualization of small compounds
   Copyright (C) 2024  Yevhen Kustovskiy

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. """

import os
import glfw
import glfw.GLFW as GLFW_CONSTANTS
from OpenGL.GLUT import *
from rdkit.Chem import MolFromMolFile, Kekulize

#from PenguinMol3D.general.renderer import Renderer
#from PenguinMol3D.objects.camera import Camera
#from PenguinMol3D.objects.light.ambient import Ambient
#from PenguinMol3D.objects.light.directional import Directional
#from PenguinMol3D.objects.light.point import Point
#from PenguinMol3D.objects.mol_3d import Mol3D
#from PenguinMol3D.objects.molecular_scene import MolecularScene
#from PenguinMol3D.objects.trackball import Trackball
from PenguinMol3D.general.renderer import Renderer
from PenguinMol3D.objects.camera import Camera
from PenguinMol3D.objects.light.directional import Directional
from PenguinMol3D.objects.mol_3d import Mol3D
from PenguinMol3D.objects.molecular_scene import MolecularScene
from PenguinMol3D.objects.trackball import Trackball
from PenguinMol3D.materials import (
    phong_material,
    pbr_material
)

"""Rendering rough surface using Blinn-Phong Lighting Model"""
class CustomPhongMaterial(phong_material.PhongMaterial):
    def __init__(self, properties: dict = {}):
        phong_material.PhongMaterial.__init__(self, properties=properties)
        self.uniforms["shininess"].data = 15.0

"""Rendering rough surface using Physically Based Lighting Model"""
class CustomPBRMaterial(pbr_material.PBRMaterial):
    def __init__(self, properties: dict = {}):
        pbr_material.PBRMaterial.__init__(self, properties=properties)
        """Sum of metallic and roughness should not exceed 1.0"""
        self.uniforms["metallic"].data = 0.4
        self.uniforms["roughness"].data = 0.6
        self.uniforms["ao"].data = 1.0

class GLFWScreenshotExample:
    def __init__(self, material: phong_material.PhongMaterial | pbr_material.PBRMaterial):
        self.title = "PenguinMol3D - Custom material example"
        self.width = 1000
        self.height = 800
        self.window = None
        self.running = True
        self.prev_pos = None

        self.make_glfw_window()
        self.make_scene(material)

    def make_glfw_window(self):
        """Create window and grab OpenGL context"""
        glfw.init()
        glfw.window_hint(GLFW_CONSTANTS.GLFW_CLIENT_API, GLFW_CONSTANTS.GLFW_OPENGL_API)
        glfw.window_hint(GLFW_CONSTANTS.GLFW_RESIZABLE, GLFW_CONSTANTS.GLFW_FALSE)
        glfw.window_hint(GLFW_CONSTANTS.GLFW_CONTEXT_VERSION_MAJOR, 3)
        glfw.window_hint(GLFW_CONSTANTS.GLFW_CONTEXT_VERSION_MINOR, 1)
        glfw.window_hint(GLFW_CONSTANTS.GLFW_SAMPLES, 8)
        glfw.window_hint(GLFW_CONSTANTS.GLFW_DEPTH_BITS, 24)

        self.window = glfw.create_window(self.width, self.height, self.title, None, None)

        glfw.set_key_callback(self.window, self.key_callback)
        glfw.set_cursor_pos_callback(self.window, self.mouse_cursor_callback)
        glfw.set_window_close_callback(self.window, self.close_callback)

        glfw.make_context_current(self.window)

    """GLFW specific methods to handle events"""
    def close_callback(self, window):
        self.running = False

    def mouse_cursor_callback(self, window, xpos: float, ypos: float):
        left_mouse_button = glfw.get_mouse_button(window, GLFW_CONSTANTS.GLFW_MOUSE_BUTTON_LEFT)
        if left_mouse_button == GLFW_CONSTANTS.GLFW_PRESS:
            if self.prev_pos:
                """Use trackball to convert previous and current cursor coordinates into rotational matrix"""
                rotmat = Trackball.simulate_trackball([self.width, self.height],
                                                      [xpos, ypos],
                                                      [self.prev_pos[0], self.prev_pos[1]])
                self.mol_3d.apply_transform(rotmat, local_coord=False)
            self.prev_pos = (xpos, ypos)

        right_mouse_button = glfw.get_mouse_button(window, GLFW_CONSTANTS.GLFW_MOUSE_BUTTON_RIGHT)
        if right_mouse_button == GLFW_CONSTANTS.GLFW_PRESS:
            if self.prev_pos:
                """Scale object to get effect similar to zoom in/zoom out"""
                if self.prev_pos[1] < ypos:
                    self.mol_3d.scale(1.1, local_coord=True)
                else:
                    self.mol_3d.scale(0.9, local_coord=True)
            self.prev_pos = (xpos, ypos)

        if left_mouse_button == GLFW_CONSTANTS.GLFW_RELEASE and \
                right_mouse_button == GLFW_CONSTANTS.GLFW_RELEASE:
            self.prev_pos = None

    def key_callback(self, window, key: int, scancode: int, action: int, mods):
        if key == GLFW_CONSTANTS.GLFW_KEY_ESCAPE and action == GLFW_CONSTANTS.GLFW_PRESS:
            self.running = False

        if key == GLFW_CONSTANTS.GLFW_KEY_S and action == GLFW_CONSTANTS.GLFW_PRESS:
            self.renderer.save_frame(self.width, self.height, "penguinone.png")

    def make_scene(self, material: phong_material.PhongMaterial | pbr_material.PBRMaterial):
        """Create the Renderer object and setup the Scene"""
        self.renderer = Renderer(clear_color=[1., 1., 1.])
        self.scene = MolecularScene()
        self.camera = Camera(angle_of_view=60,
                             aspect_ratio=self.width / self.height,
                             far=1000)

        """Load molecular data from file using one of the RDKit functions"""
        mol_rdkit = MolFromMolFile(os.path.join(os.path.dirname(os.path.abspath(__file__)), "penguinone.sdf"),
                                   sanitize=True,
                                   removeHs=False)

        Kekulize(mol_rdkit, clearAromaticFlags=True)

        """Pass rdkit Mol object and custom material to Mol3D constructor"""
        self.mol_3d = Mol3D(mol_rdkit,
                            material_type=material,
                            color_scale=1.0)
        self.scene.add_molecule(self.mol_3d)
        self.camera.set_bb_based_position(self.mol_3d.bounding_box)

        dl0, dl1, dl2 = None, None, None
        if material == CustomPhongMaterial:
            dl0 = Directional(color=[3.0, 3.0, 3.0])
            dl1 = Directional(color=[1.5, 1.5, 1.5])
            dl2 = Directional(color=[3.0, 3.0, 3.0])
        else:
            dl0 = Directional(color=[300.0, 300.0, 300.0])
            dl1 = Directional(color=[150.0, 150.0, 150.0])
            dl2 = Directional(color=[300.0, 300.0, 300.0])

        dl0.set_position([8.,8.,8])
        dl0.look_at(self.scene.get_position())
        self.scene.add_child(dl0)

        dl1.set_position(self.camera.get_position())
        dl1.look_at(self.scene.get_position())
        self.scene.add_child(dl1)

        dl2.set_position([0., 8., 0.])
        dl2.look_at(self.scene.get_position())
        self.scene.add_child(dl2)

    def run_application(self):
        """Run application event loop until user decides to quit"""
        while self.running:
            self.renderer.render(self.scene, self.camera)
            glfw.swap_buffers(self.window)
            glfw.poll_events()

        glfw.destroy_window(self.window)
        glfw.terminate()

if __name__ == "__main__":
    #app = GLFWScreenshotExample(CustomPhongMaterial)
    app = GLFWScreenshotExample(CustomPBRMaterial)
    app.run_application()

